from .core import ModelImporter  # noqa
from .fields import (  # noqa
    CachedChoiceField,
    DateTimeParserField,
    FlatRelatedField,
    JSONField,
    PreloadedChoiceField
)
from .forms import ImporterModelForm  # noqa
from .parsers import BaseImportParser, TablibCSVImportParser, TablibXLSXImportParser  # noqa
from .widgets import CompositeLookupWidget  # noqa

__version__ = '0.4.6'



from django.db import transaction
from django.db.models.fields import NOT_PROVIDED
from django.forms import modelform_factory

from .caches import SimpleDictCache


class ModelImporter:
    """ A base class which parses and processes a CSV import, and handles the priming of any required caches. """
    def __init__(self, modelimportformclass):
        """
        @param modelimportformclass The ImporterModelForm class (which extends a simple ModelForm)
        """
        self.instances = []
        self.errors = []
        self.modelimportformclass = modelimportformclass
        self.model = modelimportformclass.Meta.model
        self.update_cache = None
        self.update_queryset = None

    def get_for_update(self, pk):
        return self.update_cache[pk] if self.update_cache else self.update_queryset.get(pk=pk)

    def get_valid_fields(self, headers):
        """ Return a list of valid fields for this importer, in the order they
        appear in the input file.
        """
        valid_present_fields = [field for field in headers if field in self.modelimportformclass._meta.fields]
        virtual_fields = getattr(self.modelimportformclass.Meta, 'virtual_fields', [])  # See https://github.com/uptick/django-model-import/issues/9
        return valid_present_fields + list(virtual_fields)

    def get_required_fields(self):
        fields = self.model._meta.get_fields()
        required_fields = []

        # Required means `blank` is False and `editable` is True.
        for f in fields:
            # Note - if the field doesn't have a `blank` attribute it is probably
            # a ManyToOne relation (reverse foreign key), which you probably want to ignore.
            if getattr(f, 'blank', True) is False and getattr(f, 'editable', True) is True and f.default is NOT_PROVIDED:
                required_fields.append(f.name)
        return required_fields

    def get_modelimport_form_class(self, fields):
        """ Return a modelform for use with this data.

        We use a modelform_factory to dynamically limit the fields on the import,
        otherwise the absence of a value can be taken as false for boolean fields,
        where as we want the model's default value to kick in.
        """
        klass = modelform_factory(
            self.model,
            form=self.modelimportformclass,
            fields=fields,
        )
        base_fields_to_del = set(klass.base_fields.keys()) - set(fields)
        for f in base_fields_to_del:
            del klass.base_fields[f]
        return klass


    @transaction.atomic
    def process(self, headers, rows, commit=False, allow_update=True, allow_insert=True, limit_to_queryset=None, author=None, progress_logger=None):
        """

        @param limit_to_queryset A queryset which limits the instances which can be updated, and creates a cache of the
            updatable records to improve update performance.
        """
        # Set up a cache context which will be filled by the Cached fields
        caches = SimpleDictCache()

        # Set up an "update" cache to preload any objects which might be updated
        if allow_update:
            self.update_queryset = limit_to_queryset if limit_to_queryset is not None else self.model.objects.all()
            # We only build the update_cache if limit_to_queryset is provided, with the assumption that the dataset
            # is then not too big. This may not be a valid assumption.
            # @todo Could we be smarter about the update cache, e.g. iterate through the source row PKs
            self.update_cache = {}
            if limit_to_queryset is not None:
                for obj in self.update_queryset:
                    self.update_cache[str(obj.id)] = obj

        valid_fields = self.get_valid_fields(headers)

        # Create a Form (using modelform_factory) for rows where we are doing an UPDATE
        ModelUpdateForm = self.get_modelimport_form_class(fields=valid_fields)

        # Create a Form for rows where doing an INSERT, make sure to include the required fields
        # Combine valid & required fields; preserving order of valid fields.
        form_fields = valid_fields + list(set(self.get_required_fields()) - set(valid_fields))
        ModelImportForm = self.get_modelimport_form_class(fields=form_fields)

        # Create form to pass context to the ImportResultSet
        # TODO: evaluate this, only added because of FlatRelatedField
        header_form = ModelImportForm(data={}, caches={}, author=author)
        importresult = ImportResultSet(headers=headers, header_form=header_form)

        sid = transaction.savepoint()

        # Start processing
        for i, row in enumerate(rows, start=1):
            errors = []
            instance = None
            to_be_created = row.get('id', '') == ''  # If ID is blank we are creating a new row, otherwise we are updating
            to_be_updated = not to_be_created
            import_form_class = ModelImportForm if to_be_created else ModelUpdateForm

            if to_be_created and not allow_insert:
                errors = [('', 'Creating new rows is not permitted')]
                importresult.append(i, row, errors, instance, to_be_created)
                continue

            if to_be_updated and not allow_update:
                errors = [('', 'Updating existing rows is not permitted')]
                importresult.append(i, row, errors, instance, to_be_created)
                continue

            if to_be_updated:
                try:
                    instance = self.get_for_update(row['id'])
                except (self.model.DoesNotExist, KeyError):
                    errors = [('', '%s %s cannot be updated.' % (self.model._meta.verbose_name.title(), row['id']))]

            if not errors:
                form = import_form_class(row, caches=caches, instance=instance, author=author)
                if form.is_valid():
                    instance = form.save(commit=commit)
                else:
                    # TODO: Filter out errors associated with FlatRelatedField
                    errors = list(form.errors.items())
            result_row = importresult.append(i, row, errors, instance, to_be_created)
            if progress_logger:
                progress_logger(result_row)

        if commit:
            transaction.savepoint_commit(sid)
        else:
            transaction.savepoint_rollback(sid)

        return importresult


class ImportResultSet:
    """ Hold all imported results. """
    def __init__(self, headers, header_form):
        self.results = []
        self.headers = headers
        self.header_form = header_form

    def __repr__(self):
        return f'ImportResultSet ({len(self.results)} rows, {len(self.get_errors())} errors)'

    def append(self, index, row, errors, instance, created):
        result_row = ImportResultRow(self, index, row, errors, instance, created)
        self.results.append(result_row)
        return result_row

    def get_import_headers(self):
        return self.header_form.get_headers(self.headers)

    def get_results(self):
        return self.results

    def get_errors(self):
        return [(row.linenumber, row.errors) for row in self.results if not row.is_valid()]


class ImportResultRow:
    """ Hold the result of an imported row. """
    def __init__(self, resultset, linenumber, row, errors, instance, created):
        self.resultset = resultset
        self.linenumber = linenumber
        self.row = row
        self.errors = errors
        self.instance = instance
        self.created = created

    def __repr__(self):
        valid_str = "valid" if self.is_valid() else "invalid"
        mode_str = "create" if self.created else "update"
        res = self.get_instance_values() if self.is_valid() else self.errors
        sample = str([(k, v) for k, v in self.row.items()])[:100]
        return f'{self.linenumber}. [{valid_str}] [{mode_str}] ... {sample} ... {res}'

    def get_instance_values(self):
        return self.resultset.header_form.get_instance_values(self.instance, self.resultset.get_import_headers())

    def is_valid(self):
        return len(self.errors) == 0

    def get_errors(self):
        return self.errors



import json
import re
from dateutil import parser

from django import forms
from django.forms.utils import from_current_timezone

from .widgets import JSONFieldWidget


class UseCacheMixin:
    instancecache = None

    def set_cache(self, cache):
        self.instancecache = cache


class FlatRelatedField(forms.Field):
    """ Will create the related object if it does not yet exist.

    All the magic happens in magic.py in FlatRelatedFieldFormMixin
    """
    def __init__(self, queryset, fields=[], *args, **kwargs):
        self.queryset = queryset
        # TODO: If lookup key is provided, allow using it to look up value instead of only
        # retrieving it off the object itself.
        self.model = queryset.model
        self.fields = fields
        # Required is False, because this check gets passed down to the fields on the related instance.
        return super().__init__(required=False, *args, **kwargs)


class CachedChoiceField(UseCacheMixin, forms.Field):
    """ Use a CachedChoiceField when you have a large table of choices, but
    expect the number of different values that occur to be relatively small.

    If you expect a larger number of different values, you might want to use a
    PreloadedChoiceField.
    """
    def __init__(self, queryset, to_field=None, none_if_missing=[], *args, **kwargs):
        self.queryset = queryset
        self.model = queryset.model
        self.to_field = to_field
        self.none_if_missing = none_if_missing
        return super().__init__(*args, **kwargs)

    def get_from_cache(self, value):
        return self.instancecache[value]

    def clean(self, value):
        value = super().clean(value)

        # Fast fail if no value provided
        if not value:
            return None

        # Composite lookups are fine to have blank values in them e.g. for a firstname/lastname
        # lookup, it's fine to have ('Jenny', '').
        # However, in some situations we need some fields to be set to be able to do the lookup.
        # If they are missing then the lookup is blank.
        # @todo Think about whether this should be a validation error if self.required is True
        if self.none_if_missing:
            for field_pos in self.none_if_missing:
                if not value[field_pos]:
                    return None

        # Try and get the value from the loader
        try:
            return self.get_from_cache(value)
        except self.model.DoesNotExist:
            raise forms.ValidationError(
                "No %s matching '%s'." % (self.model._meta.verbose_name.title(), value)
            )
        except self.model.MultipleObjectsReturned:
            raise forms.ValidationError(
                "Multiple %s matching '%s'. Expected just one." % (self.model._meta.verbose_name_plural.title(), value)
            )


class PreloadedChoiceField(forms.Field):
    """ This will load all the possible values for this relationship once,
    to avoid hitting the database for each relationship in the import.
    """
    def clean(self, value):
        raise NotImplementedError


class DateTimeParserField(forms.DateTimeField):
    """ A DateTime parser field that does it's best effort to understand.

        Defaults to assuming little endian when there is ambiguity:
        - XX/XX/XX -> DD/MM/YY
        - XX/XX/XXXX -> DD/MM/YYYY

        Pass in `middle_endian=True` to get:
        - XX/XX/XX -> MM/DD/YY
        - XX/XX/XXXX -> MM/DD/YYYY

        If year is passed first, will always use big endian:
        - XXXX/XX/XX -> YYYY/MM/DD
    """
    def __init__(self, middle_endian=False, *args, **kwargs):
        self.middle_endian = middle_endian
        return super().__init__(*args, **kwargs)

    def to_python(self, value):
        value = (value or '').strip()
        if value:
            try:
                dayfirst = not bool(re.match(r'^\d{4}.\d\d?.\d\d?', value)) and not self.middle_endian
                return from_current_timezone(parser.parse(value, dayfirst=dayfirst))
            except (TypeError, ValueError, OverflowError):
                raise forms.ValidationError(self.error_messages['invalid'], code='invalid')

        else:
            return None


class JSONField(forms.Field):
    """ This lets you store any fields prefixed by the field name into a JSON blob.

    For example, adding a field:
        metadata = JSONField()

    When the row is submitted with data that looks like this:
    id  name    author  metadata_rank   metadata_score
    -----------------------------------------------
        ding    bob     hello           twenty

    This field will return a JSON blob that looks like:
        {rank: "hello", score: "twenty"}
    """
    def __init__(self, **kwargs):
        kwargs['widget'] = kwargs.get('widget', JSONFieldWidget)
        kwargs['required'] = False
        kwargs['initial'] = dict
        super().__init__(**kwargs)

    def validate_json(self, value, is_serialized=False):
        # if empty
        if value is None or value == '' or value == 'null':
            value = '{}'

        # ensure valid JSON
        try:
            # convert strings to dictionaries
            if isinstance(value, str):
                dictionary = json.loads(value)

                # if serialized field, deserialize values
                if is_serialized and isinstance(dictionary, dict):
                    dictionary = dict((k, json.loads(v)) for k, v in dictionary.items())  # TODO: modify to use field's deserializer
            # if not a string we'll check at the next control if it's a dict
            else:
                dictionary = value
        except ValueError as e:
            raise forms.ValidationError(('Invalid JSON: {0}').format(e))

        # ensure is a dictionary
        if not isinstance(dictionary, dict):
            raise forms.ValidationError(('No lists or values allowed, only dictionaries'))

        # convert any non string object into string
        for key, value in dictionary.items():
            if isinstance(value, dict) or isinstance(value, list):
                dictionary[key] = json.dumps(value)
            if isinstance(value, bool) or isinstance(value, int) or isinstance(value, float):
                if not is_serialized:  # Only convert if not from serializedfield
                    dictionary[key] = str(value).lower()

        return dictionary

    def to_python(self, value):
        return self.validate_json(value)

    def render(self, name, value, attrs=None):
        # return json representation of a meaningful value
        # doesn't show anything for None, empty strings or empty dictionaries
        if value and not isinstance(value, str):
            value = json.dumps(value, sort_keys=True, indent=4)
        return super().render(name, value, attrs)



from django import forms

from .magic import CachedChoiceFieldFormMixin, FlatRelatedFieldFormMixin, JSONFieldFormMixin


class ImporterModelForm(JSONFieldFormMixin, FlatRelatedFieldFormMixin, CachedChoiceFieldFormMixin, forms.ModelForm):
    """ Extends the ModelForm to prime our caches and tweaks the validation
    routines to ensure we are not doing too many queries with our cached fields.
    """
    def __init__(self, data, caches, author=None, *args, **kwargs):
        self.caches = caches
        self.author = author
        super().__init__(data, *args, **kwargs)

    # This improves preview performance but eliminates validation on uniqueness constraints
    # def validate_unique(self):
    #     pass



import operator

from django import forms


class DisplayChoiceWidget(forms.Widget):
    """ This widget is helpful when the value being uploaded by the customer is the
    display choice, not the value. This widget will map the display choice back to the value.
    """
    choices = None
    display_to_choice_map = None

    def flip_enum(self, choices):
        return dict(zip(dict(choices).values(), dict(choices).keys()))

    def __init__(self, choices, *args, **kwargs):
        self.choices = choices
        self.display_to_choice_map = self.flip_enum(choices)
        return super().__init__(choices, *args, **kwargs)

    def value_from_datadict(self, data, files, name):
        """
        Given a dictionary of data and this widget's name, return the value
        of this widget or None if it's not provided.
        """
        val = data.get(name)
        return self.display_to_choice_map.get(val)

    def format_value(self, value):
        return self.choices.get(value)


class CompositeLookupWidget(forms.Widget):
    def __init__(self, source, *args, **kwargs):
        self.source = source
        super().__init__(*args, **kwargs)

    def value_from_datadict(self, data, files, name):
        getter = operator.itemgetter(*self.source)
        try:
            return getter(data)
        except KeyError:
            pass

    def value_omitted_from_data(self, data, files, name):
        for field_name in self.source:
            if field_name not in data:
                return True
        return False


class JSONFieldWidget(forms.Widget):
    template_name = 'django/forms/widgets/textarea.html'

    def render(self, name, value, attrs=None, renderer=None):
        return ''

    def value_omitted_from_data(self, data, files, name):
        return not any([key.startswith(name) for key in data.keys()])

    def value_from_datadict(self, data, files, name):
        extra_fields = {}
        for f in data.keys():
            if f.startswith(name):
                new_field = f[len(name) + 1:]
                extra_fields[new_field] = data[f]
        return extra_fields



class BaseImportParser:
    def __init__(self, modelvalidator):
        """ We provide the modelvalidator to get some Meta information about
        valid fields, and any soft headings.
        """
        self.modelvalidator = modelvalidator

    def get_soft_headings(self):
        # Soft headings are used to provide similar heading suggestions
        # and look like this: {the field name: [list of other possible names] }
        # eg.
        #
        # soft_headings = {
        #   'type': ['Asset Type'],
        # }
        header_map = {}
        if hasattr(self.modelvalidator, 'ImporterMeta'):
            if hasattr(self.modelvalidator.ImporterMeta, 'soft_headings'):
                importer_softheadings = self.modelvalidator.ImporterMeta.soft_headings

                for renameto in importer_softheadings:  # new column name
                    for renamefrom in importer_softheadings[renameto]:  # old column name
                        header_map[renamefrom.lower()] = renameto.lower()
        return header_map

    def parse(self, data):
        """ Parsers should return a tuple containing (headings, data)

        They should also take a dictionary of soft_headings which map
        similar names to actual headings.
        """
        raise NotImplementedError


class TablibBaseImportParser(BaseImportParser):
    def __init__(self, *args, **kwargs):
        # Inline import, so tablib is only grabbed if/when this Parser is instanciated.
        from tablib import Dataset
        self.dataset_class = Dataset
        super().__init__(*args, **kwargs)


class TablibCSVImportParser(TablibBaseImportParser):
    def parse(self, data):
        dataset = self.dataset_class()
        dataset.csv = data

        header_map = self.get_soft_headings()

        # Make all our headings lowercase and sub in soft headings
        for col_id, header in enumerate(dataset.headers):  # replace it in headers if found
            header_name = header.strip().lower()
            dataset.headers[col_id] = header_name
            if header_name in header_map.keys():
                dataset.headers[col_id] = header_map[header_name]

        return (dataset.headers, dataset.dict)


class TablibXLSXImportParser(TablibBaseImportParser):
    def parse(self, data):
        dataset = self.dataset_class()
        # TODO: This does not currently work, as dataset.xlsx cannot be set.
        # http://docs.python-tablib.org/en/latest/api/#tablib.Dataset.xlsx
        # We can wait for it to be supported, or in the meantime, use this converter:
        # https://github.com/dilshod/xlsx2csv
        dataset.xlsx = data  # CANNOT SET
        return (dataset.headers, dataset.dict)



from django.core.exceptions import ValidationError
from django.forms import FileField

from .fields import CachedChoiceField, FlatRelatedField, JSONField, UseCacheMixin
from .loaders import CachedInstanceLoader

""" These mixins hold all the code that relates to our special fields (flat related, json, cached choice)
that just doesn't work without access to the form instance. """


class FlatRelatedFieldFormMixin:
    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        self.flat_related_mapping = {}
        flat_related = {}

        for field, fieldinstance in self.fields.items():
            # For each FlatRelatedField, save a mapping back to the field.
            if isinstance(fieldinstance, FlatRelatedField):
                flat_related[field] = {}
                for f in fieldinstance.fields.keys():
                    self.flat_related_mapping[f] = field

        # Tinker with data to combine flat fields into related objects.
        new_data = self.data.copy()
        for field, value in self.data.items():
            if field in self.flat_related_mapping:
                flat_related[self.flat_related_mapping[field]][field] = value
                del new_data[field]

        self.flat_data = self.data
        self.data = new_data

        for field, values in flat_related.items():
            mapped_values = dict((self.fields[field].fields[k]['to_field'], v) for k, v in values.items())
            # Get or create the related instance.
            if getattr(self.instance, field + '_id') is None:
                instance = self.fields[field].model(**mapped_values)
            else:
                instance = getattr(self.instance, field)
                for attr, value in mapped_values.items():
                    setattr(instance, attr, value)

            instance.save()  # @todo this gets fired during preview :( this whole thing needs to go
            self.data[field] = instance

    def get_headers(self, given_headers=None):
        headers = []
        for field, fieldinstance in self.fields.items():
            if isinstance(fieldinstance, FlatRelatedField):
                headers.extend(f for f in fieldinstance.fields.keys() if given_headers is None or f in given_headers)
            else:
                headers.append(field)
        return headers

    def get_instance_values(self, instance, headers):
        instance_values = []
        for header in headers:
            if header in self.flat_related_mapping:
                rel_field_name = self.flat_related_mapping[header]
                rel = getattr(instance, rel_field_name)
                instance_values.append(getattr(rel, self.fields[rel_field_name].fields[header]['to_field']))
            else:
                try:
                    instance_values.append(getattr(instance, header))
                except ValueError:
                    instance_values.append('')  # trying to access an m2m is not allowed before it has been saved
                except AttributeError:
                    instance_values.append('')  # trying to access a field that doesn't exist on the model definition, should we check for the field in _meta.exclude?
        return instance_values

    # TODO:
    # def full_clean(self):
    #     """ Validate that required fields in FlatRelated fields have been provided. """
    #     super().full_clean()
    #     for field, fieldinstance in self.fields.items():
    #         if isinstance(fieldinstance, FlatRelatedField):
    #             for f in fieldinstance.fields:
    #                 import pdb; pdb.set_trace()
    #                 pass


class CachedChoiceFieldFormMixin:
    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        for field, fieldinstance in self.fields.items():
            # For each CachedInstanceLoader, prime the cache.
            if isinstance(fieldinstance, UseCacheMixin):
                if field not in self.caches:
                    self.caches[field] = CachedInstanceLoader(fieldinstance.queryset, fieldinstance.to_field)
                fieldinstance.set_cache(self.caches[field])

    def _get_validation_exclusions(self):
        """ We need to exclude any CachedChoiceFields from validation, as this
        causes a m * n queries where m is the number of relations, n is rows.
        """
        exclude = []
        for field, fieldinstance in self.fields.items():
            if isinstance(fieldinstance, CachedChoiceField):
                exclude.append(field)
        return exclude


class JSONFieldFormMixin:
    def _clean_fields(self):
        for name, field in self.fields.items():
            # value_from_datadict() gets the data from the data dictionaries.
            # Each widget type knows how to retrieve its own data, because some
            # widgets split data over several HTML fields.
            if field.disabled:
                value = self.get_initial_for_field(field, name)
            else:
                value = field.widget.value_from_datadict(self.data, self.files, self.add_prefix(name))
            try:
                if isinstance(field, FileField):
                    initial = self.get_initial_for_field(field, name)
                    value = field.clean(value, initial)
                # PATCH
                if isinstance(field, JSONField):
                    initial = getattr(self.instance, name)
                    value = field.clean(value)
                    value = dict(initial, **value)  # this is the secret sauce.
                # ENDPATCH
                else:
                    value = field.clean(value)
                self.cleaned_data[name] = value
                if hasattr(self, 'clean_%s' % name):
                    value = getattr(self, 'clean_%s' % name)()
                    self.cleaned_data[name] = value
            except ValidationError as e:
                self.add_error(name, e)



class CachedInstanceLoader(dict):
    """ A clever cache that queries the database for any missing objects.

    If there's an error, it's only raised against the first item that causes it, then it's
    cached for extra speed.
    """
    def __init__(self, queryset, to_field, *args, **kwargs):
        self.queryset = queryset
        self.model = queryset.model
        self.to_field = to_field
        self.multifield = isinstance(to_field, list) or isinstance(to_field, tuple)

    def __missing__(self, value):
        if self.multifield:
            params = dict(zip(self.to_field, value))
        else:
            params = {self.to_field: value}

        try:
            self[value] = inst = self.queryset.get(**params)
        except self.model.DoesNotExist:
            self[value] = inst = None  # Further warnings will be surpressed
            raise
        except self.model.MultipleObjectsReturned:
            self[value] = inst = None  # Further warnings will be surpressed
            raise
        return inst



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djangomodelimport']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.0.0,<4.0.0',
 'python-dateutil>=2.7.0,<3.0.0',
 'tablib>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'django-model-import',
    'version': '0.4.10',
    'description': '',
    'long_description': '# django-model-import\n\n[![PyPI version](https://badge.fury.io/py/django-model-import.svg)](https://badge.fury.io/py/django-model-import)\n[![Build Status](https://travis-ci.org/uptick/django-model-import.svg?branch=master)](https://travis-ci.org/uptick/django-model-import)\n\nDjango Model Import is a light weight CSV importer built for speed.\n\nIt uses a standard Django `ModelForm` to parse each row, giving you a familiar API to work with\nfor data validation and model instantiation. In most cases, if you already have a `ModelForm`\nfor the `ContentType` you are importing you do not need to create an import specific form.\n\nTo present feedback to the end-user running the import you can easily generate a preview\nof the imported data by toggling the `commit` parameter.\n\nIt also provides some import optimized fields for ForeignKey\'s, allowing preloading all\npossible values, or caching each lookup as it occurs, or looking up a model where multiple\nfields are needed to uniquely identify a resource.\n\n\n## Installation\n\n```bash\npoetry add django-model-import\n```\n\n\n## Quickstart\n\n```python\nimport djangomodelimport\n\nclass BookImporter(djangomodelimport.ImporterModelForm):\n    name = forms.CharField()\n    author = CachedChoiceField(queryset=Author.objects.all(), to_field=\'name\')\n\n    class Meta:\n        model = Book\n        fields = (\n            \'name\',\n            \'author\',\n        )\n\nwith default_storage.open(\'books.csv\', \'rb\') as fh:\n    data = fh.read().decode("utf-8")\n\n# Use tablib\nparser = djangomodelimport.TablibCSVImportParser(BookImporter)\nheaders, rows = parser.parse(data)\n\n# Process\nimporter = djangomodelimport.ModelImporter(BookImporter)\npreview = importer.process(headers, rows, commit=False)\nerrors = preview.get_errors()\n\nif errors:\n    print(errors)\n\nimportresult = importer.process(headers, rows, commit=True)\nfor result in importresult.get_results():\n    print(result.instance)\n```\n\n\n## Composite key lookups\n\nOften a relationship cannot be referenced via a single unique string. For this we can use\na `CachedChoiceField` with a `CompositeLookupWidget`. The widget looks for the values\nunder the `type` and `variant` columns in the source CSV, and does a unique lookup\nwith the field names specified in `to_field`, e.g. `queryset.get(type__name=type, name=variant)`.\n\nThe results of each `get` are cached internally for the remainder of the import minimising\nany database access.\n\n```python\nclass AssetImporter(ImporterModelForm):\n    site = djangomodelimport.CachedChoiceField(queryset=Site.objects.active(), to_field=\'ref\')\n    type = djangomodelimport.CachedChoiceField(queryset=AssetType.objects.filter(is_active=True), to_field=\'name\')\n    type_variant = djangomodelimport.CachedChoiceField(\n        queryset=InspectionItemTypeVariant.objects.filter(is_active=True),\n        required=False,\n        widget=djangomodelimport.CompositeLookupWidget(source=(\'type\', \'variant\')),\n        to_field=(\'type__name\', \'name\'),\n    )\n    contractor = djangomodelimport.CachedChoiceField(queryset=Contractor.objects.active(), to_field=\'name\')\n```\n\n\n## Flat related fields\n\nOften you\'ll have a OneToOneField or just a ForeignKey to another model, but you want to be able to\ncreate/update that other model via this one. You can flatten all of the related model\'s fields onto\nthis importer using `FlatRelatedField`.\n\n```python\nclass ClientImporter(ImporterModelForm):\n    primary_contact = FlatRelatedField(\n        queryset=ContactDetails.objects.all(),\n        fields={\n            \'contact_name\': {\'to_field\': \'name\', \'required\': True},\n            \'email\': {\'to_field\': \'email\'},\n            \'email_cc\': {\'to_field\': \'email_cc\'},\n            \'mobile\': {\'to_field\': \'mobile\'},\n            \'phone_bh\': {\'to_field\': \'phone_bh\'},\n            \'phone_ah\': {\'to_field\': \'phone_ah\'},\n            \'fax\': {\'to_field\': \'fax\'},\n        },\n    )\n\n    class Meta:\n        model = Client\n        fields = (\n            \'name\',\n            \'ref\',\n            \'is_active\',\n            \'account\',\n\n            \'primary_contact\',\n        )\n```\n\n## Tests\nRun tests with `python example/manage.py test testapp`\n',
    'author': 'Aidan Lister',
    'author_email': 'aidan@uptickhq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uptick/django-model-import',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
